/*
** Module   : ComboBoxWidget.java
** Abstract : server side combo box widget implementation
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23483 Created initial version.
** 002 SIY 20050209   @24433 Major refactoring.
** 003 SIY 20050209   @24494 Added missing setInnerLines method.
** 004 GES 20060803   @28368 Some cleanup and refactoring.
** 005 GES 20080320   @37583 Added TYPE attribute.
** 006 SIY 20111101          Updated imports.
** 007 CA  20130221          Refactored usage of TYPE attribute.
** 008 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 009 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 010 CA  20131115          Changes related to properly instantiate dynamic resources.
** 011 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 012 SVL 20141102          Added MAX-CHARS and AUTO-ZAP attributes.
** 013 GES 20141124          Added INNER-LINES interface, NUM-ITEMS attribute and the
**                           SET-SELECTION() method support.
** 014 IAS 20150123          Implemented DELETE() method
** 015 IAS 20150209          Fixed behavior to be compatible with 4GL.
** 016 IAS 20150225          Fixed support for the LIST-ITEMS and LIST-ITEM-PAIRS attributes
** 017 IAS 20150604          Added support for the REPLACE method, text selection, 
**                           AUTO-ZAP, DISABLE-AUTO-ZAP, CURSOR-OFFSET, NUM-ITEMS 
**                           and INNER-LINES attribute, realize() operation .
*/ 

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Server side combo box widget implementation.
 */
@LegacyResource(resource = LegacyResource.COMBO_BOX)
public class ComboBoxWidget
extends ControlSetEntity<ComboBoxConfig>
implements AutoZapElement,
           MaxCharsElement,
           NumItems,
           SortedElement,
           InnerLines,
           SelectableText
{
   /**
    * Default constructor.
    */
   public ComboBoxWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    */
   public ComboBoxWidget(boolean dynamic)
   {
      super(dynamic, new ComboBoxConfig());
   }
   
   /**
    * Set DROP-DOWN-LIST mode for the widget. Has no effect in character mode
    * interface.
    * 
    * @param   b
    *          <code>true</code> enables mode, <code>false</code> is
    *          ignored.
    */
   public void setModeDropDownList(boolean b)
   {
      // do nothing
   }

   /**
    * Set DROP-DOWN mode for the widget. Has no effect in character mode
    * interface.
    * 
    * @param   b
    *          <code>true</code> enables mode, <code>false</code> is
    *          ignored.
    */
   public void setModeDropDown(boolean b)
   {
      // do nothing
   }
   
   /**
    * Set SIMPLE mode for the widget. Has no effect in character mode
    * interface.
    * 
    * @param   b
    *          <code>true</code> enables mode, <code>false</code> is
    *          ignored.
    */
   public void setModeSimple(boolean b)
   {
      // do nothing
   }

   /**
    * Get the NUM-ITEMS attribute.
    *
    * @return   The numebr of items in the list.
    */
   @Override
   public integer getNumItems()
   {
      return new integer(config.items == null ? 0 : config.items.length);
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int64 start, int64 end)
   {
      return setSelection(start == null || start.isUnknown() ? 0 : start.intValue(), 
            end == null || end.isUnknown() ? 0 : end.intValue());
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int64 start, int end)
   {
      return setSelection(start == null || start.isUnknown() ? 0 : start.intValue(), end);
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int start, int64 end)
   {
      return setSelection(start, end == null || end.isUnknown() ? 0 : end.intValue());
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int start, int end)
   {
      if (!canAccess("SET-SELECTION"))
      {
         return new logical();
      }

      if (start == 0 || end == 0)
      {
         ErrorManager.recordOrShowError(4057, 
               String.format(
               "Attribute SET-SELECTION for the %s %s has an invalid value of 0", 
               type(), widgetName()),
               false);
         return new logical();
      }
      int s1 = Math.min(start, end);
      int s2 = Math.max(start,  end);
      if (s1 < 0 || s2 > config.widthChars - 4)
      {
         return new logical(false);
      }

      config.selectionStart = s1;
      config.selectionEnd = s2;
      config.selectionActive = true;
      pushScreenDefinition();
      return new logical(true);
   }
   
   /**
    * Get SELECTION-TEXT attribute.
    *
    * @return  SELECTION-TEXT value   
    */
   public character getSelectionText()
   {
      if (!canAccess("SELECTION-TEXT"))
      {
         return new character();
      }

      if (!config.selectionActive)
      {
         return new character("");
      }
      String s = "";
      if (config.items != null && 
            config.selectedIndex >= 0 && config.selectedIndex < config.items.length)
      {
         ControlSetItem item = config.items[config.selectedIndex];
         s = item.getLabel().toStringMessage();
      }
      
      s += StringHelper.repeatChar(' ', (int)config.widthChars -4 - s.length());
      return new character(
            config.selectionEnd + 1 >= (int)config.widthChars - 4 ?
               s.substring(config.selectionStart) :
               s.substring(config.selectionStart, config.selectionEnd + 1)
            );
   }

   /**
    * Get SELECTION-END attribute.
    *
    * @return  SELECTION-END value   
    */
   public integer getSelectionEnd()
   {
      if (!canAccess("SELECTION-END"))
      {
         return new integer();
      }

      if (!config.selectionActive)
      {
         return new integer();
      }
      return new integer(config.selectionEnd);
   }

   /**
    * Get SELECTION-START attribute.
    *
    * @return  SELECTION-START value   
    */
   public integer getSelectionStart()
   {
      if (!canAccess("SELECTION-START"))
      {
         return new integer();
      }

      if (!config.selectionActive)
      {
         return new integer();
      }
      return new integer(config.selectionStart);
   }

   /**
    * Get TEXT-SELECTED attribute.
    *
    * @return  TEXT-SELECTED value   
    */
   @Override
   public logical isTextSelected()
   {
      if (!canAccess("TEXT-SELECTION"))
      {
         return new logical();
      }
      return new logical(config.selectionActive);
   }

   /**
    * Clears text selection
    */
   @Override
   public logical clearSelection()
   {
      boolean wasSelection = config.selectionActive;
      config.selectionActive = false;
      pushScreenDefinition();
      return new logical(wasSelection);
   }

   /**
    * Gets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @return   The number of lines of text.
    */
   @Override
   public integer getInnerLines()
   {
      if (!config.realized)
      {
         return new integer(0);
      }
      return new integer(config.innerLines);
   }
   
   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   @Override
   public void setInnerLines(double lines)
   {
      int value = (int) lines;
      if (value < 0)
      {
         ErrorManager.recordOrShowError(4082, 
               String.format(
               "Attribute INNER-LINES must be greater than or equal to 0 on %s %s", 
               type(), widgetName()),
               false);
         return;
      }

      config.innerLinesAssignments++;
      if (!config.realized)
      {
         config.innerLines = value;
      }
      else 
      {
         config.innerLines = Math.min(value, 
               config.innerLinesLimit - config.innerLinesAssignments);
      }
      if (frame != null)
      {
         config.heightChars = config.innerLines + 2; 
      }
      pushScreenDefinition();
   }

   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   @Override
   public void setInnerLines(NumberType lines)
   {
      if (lines == null || lines.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute INNER-LINES on %s %s", 
               type(), widgetName()), false);
         return;
      }
   
      setInnerLines(lines.intValue());
   }

   /**
    * Gets the HEIGHT-CHARS writable attribute.
    *
    * @return   The current value of the HEIGHT-CHARS attribute.
    */
   @Override
   public decimal getHeightChars()
   {
      if (!canAccess("HEIGHT"))
      {
         return new decimal(1);
      }
      return frame == null ? new decimal(1) : super.getHeightChars();
   }

   /**
    * Sets the DISABLE-AUTO-ZAP option.
    * 
    * @param    disableAutoZap
    *           <code>true</code> if the option is set to ON.
    */
   @Override
   public void setDisableAutoZap(logical disableAutoZap)
   {
      notSettable("DISABLE-AUTO-ZAP");
   }

   /**
    * Sets the DISABLE-AUTO-ZAP option.
    * 
    * @param    disableAutoZap
    *           <code>true</code> if the option is set to ON.
    */
   @Override
   public void setDisableAutoZap(boolean disableAutoZap)
   {
      notSettable("DISABLE-AUTO-ZAP");
   }

   /**
    * Get value of the DISABLE-AUTO-ZAP attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isDisableAutoZap()
   {
      notQueryable("DISABLE-AUTO-ZAP");
      return new logical();
   }

   /**
    * Set AUTO-ZAP attribute.
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setAutoZap(logical value)
   {
      notSettable("AUTO-ZAP");
   }

   /**
    * Set AUTO-ZAP attribute.
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setAutoZap(boolean value)
   {
      notSettable("AUTO-ZAP");
   }

   /**
    * Get value of the AUTO-ZAP attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isAutoZap()
   {
      notQueryable("AUTO-ZAP");
      return new logical();
   }

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute 
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @Override
   public void setCursorOffset(double offset)
   {
      notSettable("CURSOR-OFFSET");
   }

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute 
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @Override
   public void setCursorOffset(NumberType offset)
   {
      notSettable("CURSOR-OFFSET");
   }

   /**
    * Obtains the index position of the cursor in a text editing widget. This corresponds to
    * the CURSOR-OFFSET attribute in BROWSE columns (only when the cell is FILL-IN or COMBO-BOX),
    * COMBO-BOX, EDITOR or FILL-IN widgets.
    * 
    * @return   The character index at which the cursor is placed.
    */
   @Override
   public integer getCursorOffset()
   {
      notQueryable("CURSOR-OFFSET");
      return new integer();
   }
   /**
    * Set value of the MAX-CHARS attribute.
    *
    * @param    maxChars
    *           Value of the attribute.
    */
   @Override
   public void setMaxChars(double maxChars)
   {
      setMaxChars(new decimal(maxChars));
   }

   /**
    * Set value of the MAX-CHARS attribute.
    *
    * @param    maxChars
    *           Value of the attribute.
    */
   @Override
   public void setMaxChars(NumberType maxChars)
   {
      if (maxChars.isUnknown())
         return;

      config.maxChars = maxChars.intValue();
      pushScreenDefinition();
   }

   /**
    * Get value of the MAX-CHARS attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public integer getMaxChars()
   {
      return new integer(config.maxChars);
   }

   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param itemList
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(character itemList)
   {
      return delete(itemList, COMPARE_VALUE_IGNORE_CASE, true);
   }

   /**
    * Finish initialization of the combo box widget. 
    */
   public void finishSetup()
   {
      if (frame == null)
         return;
 
      FrameConfig frameConfig = frame.getFrameWidget().config();
      config.innerLinesLimit = LogicalTerminal.getScreenLines().intValue() - 
            ((int)frameConfig.row + 1) + 2;
      if (frameConfig.box)
         config.innerLinesLimit--;

      /*
       *  This value of innerLinesLimitDelta is valid in the case when there is only
       *  one frame F containing the widget W.
       *  When there are n frames
       *  Fn -> ... -> F -> W
       *  then the value of innerLinesLimitDelta is
       *  SUM(Fi:ROW + (if Fi:BOX then 1 else 0) - 1 | i = 1,...,n) + W:ROW.
       *  
       *  At this moment nested frames are not supported.   
       */
      config.innerLinesLimitDelta = (int)config.row;
      
      config.innerLines = Math.max(config.items == null ? 0 : config.items.length, 
            Math.min(config.innerLines, config.innerLinesLimit));
      config.innerLinesAssignments = 0;
   }

   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param itemList
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   public logical delete(String itemList)
   {
      return delete(itemList, COMPARE_VALUE_IGNORE_CASE, true);
   }
   
   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   @Override
   public void setFrame(FrameWidget frame)
   {
      super.setFrame(frame);
      if (config.innerLines < 0)
      {
         config.innerLines = 2; 
      }
      config.heightChars = 1;
   }

   /**
    * Remove trailing whitespaces
    * 
    * @param value
    *    the string to be truncated
    *    
    * @return the truncated argument
    */
   @Override
   protected String truncateValue(String value)
   {
         return RTRIM.matcher(value).replaceAll("");
   }

   /**
    * Remove trailing whitespaces if applicable
    * 
    * @param value
    *    the BaseDataType to be truncated
    *    
    * @return
    *    the truncated argument if the argument is of type character
    *    (nothing is required for non-character types)
    */
   @Override
   protected BaseDataType truncateValue(BaseDataType value)
   {
      if (value instanceof character)
      {
         return new character(truncateValue(value.toStringMessage()));
      }
      return value;
   }

   /**
    * Create new ControlSetItem
    * 
    * @param label
    *       item' label
    * @param value
    *       item's value
    *       
    * @return 
    *    new instance of the ControlSetItem
    */
   @Override
   protected ControlSetItem controlSetItem(character label, BaseDataType value) 
   {
      return new ControlSetItem(config.itemId++, 
            label.isUnknown() ? new character(" ") : (character)label.duplicate(), 
            truncateValue(value.duplicate())); 
   }
   
   /**
    * Truncate values of the provided items and remove invalid ones.
    * 
    * @param items
    *    items to be processed
    *    
    * @return
    *    valid items with truncated values
    */
   @Override
   protected ControlSetItem[] truncateValues(ControlSetItem[] items)
   {
      List<ControlSetItem> list = new ArrayList<>(items.length);
      for (ControlSetItem item : items)
      {
         if (invalid(item.getLabel(), item.getValue()))
         {
            continue;
         }
         list.add(controlSetItem(item.getLabel(), item.getValue()));
      }
      return list.toArray(new ControlSetItem[list.size()]);
   }
   
   /**
    * 
    * Checks if a value should be ignored (UNKNOWN for COMBO-BOX)
    *  
    * @param 
    *    value - value to be checked
    *    
    * @return <code>true</code> if the value should be ignored
    */
   @Override
   protected boolean ignore(BaseDataType value)
   {
      return value != null && "\"?\"".equals(value.toStringExport()); 
   }

   /**
    * Checks if a value should be ignored ("?" for the COMBO-BOX)
    *  
    * @param value
    *          value to be checked
    *    
    * @return <Code>true</code> if the value should be ignored
    */
   @Override
   protected boolean ignore(String value)
   {
      return "?".equals(value); 
   }
}
