/*
** Module   : RadioSetWidget.java
** Abstract : server side radioset widget implementation
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23494 Created initial version.
** 002 SIY 20050209   @24428 Major refactoring.
** 003 GES 20060803   @28382 Some cleanup and refactoring.
** 004 GES 20080320   @37593 Added TYPE attribute.
** 018 SIY 20111101          Updated imports.
** 019 CA  20130221          Refactored usage of TYPE attribute.
** 020 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 021 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 022 CA  20131115          Changes related to properly instantiate dynamic resources.
** 023 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 025 IAS 20150123          Implemented DELETE() method
** 026 IAS 20150202          Implemented ADD_LAST() method, fixed ADD-FIRST()
** 027 IAS 20150304          Fixed getRadioButtons() & setItems() methods
** 028 CA  20150310          Changes related to ControlSetItems.setItems API change.
** 029 IAS 20150312          Implemented ENABLE/DISABLE method and NUM-BUTTONS attribute.
** 030 HC  20150323          Removed business logic from config classes
** 031 IAS 20150604          Added support for the REPLACE method and realize() operation .
*/ 

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.util.*;

/**
 * Server side radioset widget implementation.
 */
@LegacyResource(resource = LegacyResource.RADIO_SET)
public class RadioSetWidget
extends ControlSetEntity<RadioSetConfig>
implements RadioSetInterface
{
   /**
    * Default constructor.
    */
   public RadioSetWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public RadioSetWidget(boolean dynamic)
   {
      super(dynamic, new RadioSetConfig());
   }

   /**
    * Resets the value of the HORIZONTAL attribute for radio-set.
    * 
    * @param   value
    *          The new value of the HORIZONTAL attribute.
    */
   @Override
   public void setHorizontal(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, String.format(
               "Unable to assign UNKNOWN value to attribute HORIZONTAL on %s %s", type(),
                  widgetName()), false);
      }
      setHorizontal(value.booleanValue());
   }

   /**
    * Resets the value of the HORIZONTAL attribute for radio-set.
    * 
    * @param   value
    *          The new value of the HORIZONTAL attribute.
    */
   @Override
   public void setHorizontal(boolean value)
   {
      if (!config.realized)
      {
         config.horizontal = value;
         resetSize(config.items);
      }
      else {
         ErrorManager.recordOrShowError(4053, String.format(
               "Unable to set HORIZONTAL because the %s %s has been realised", type(),
                  widgetName()), false);
      }
   }

   /**
    * Get the value of the HORIZONTAL attribute for radio-set.
    * 
    * @return
    *          The new value of the HORIZONTAL attribute.
    */
   @Override
   public logical isHorizontal()
   {
      return new logical(config.horizontal);
   }
   
   /**
    * Get the value of the EXPAND attribute for radio-set.
    * 
    * @return
    *          The value of the EXPAND attribute.
    */
   @Override
   public logical isExpand()
   {
      return new logical(config.expand);
   }

   /**
    * Resets the value of the EXPAND attribute for radio-set.
    * 
    * @param   value
    *          The new value of the EXPAND attribute.
    */
   @Override
   public void setExpand(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, String.format(
               "Unable to assign UNKNOWN value to attribute EXPAND on %s %s", type(),
                  widgetName()), false);
      }
      setExpand(value.booleanValue());
   }

   /**
    * Resets the value of the EXPAND attribute for radio-set.
    * 
    * @param   value
    *          The new value of the EXPAND attribute.
    */
   @Override
   public void setExpand(boolean value)
   {
      if (!config.realized)
      {
         config.expand = value;
      }
      else {
         ErrorManager.recordOrShowError(4053, String.format(
               "Unable to set EXPAND because the %s %s has been realised", type(),
                  widgetName()), false);
      }
   }

   
   /**
    * Get RADIO-BUTTONS attribute.
    * 
    * @return  Current value of the attribute.
    */
   @Override
   public character getRadioButtons()
   {
      if (config.items == null || config.items.length == 0)
      {
         ErrorManager.recordOrShowError(4094, 
               String.format(
                  "Attribute RADIO-BUTTONS is not queryable on %s %s.  It has no buttons", 
                  type(), widgetName()), 
               false);
         return new character();
      }
      return getListItemPairs();
   }
   
   /**
    * Set RADIO-BUTTONS attribute. The parameter contains string
    * representation of all radio button labels and values separated by
    * delimiter (comma is the default value).
    * <p>
    * Example: <code>Label1,value1,Label2,value2,..., LabelN,valueN</code>
    * 
    * @param   s
    *          New value of the attribute.
    */
   @Override
   public void setRadioButtons(String s)
   {
      setRadioButtons(new character(s));
   }

   /**
    * Set RADIO-BUTTONS attribute. The parameter contains string
    * representation of all radio button labels and values separated by
    * delimiter (comma is the default value).
    * <p>
    * Example: <code>Label1,value1,Label2,value2,..., LabelN,valueN</code>
    * 
    * @param   c
    *          New value of the attribute.
    */
   @Override
   public void setRadioButtons(character c)
   {
      setListItemPairs(c);
   }

   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param item
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(character item)
   {
      if (!canAccess("DELETE"))
      {
         return new logical(false);
      }
     if (item == null || item.isUnknown())
      {
         invalidDeleteArg();
         return new logical(false);
      }
      return delete(item.toStringMessage());
   }

   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param item
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(String item)
   {
      if (!canAccess("DELETE"))
      {
         return new logical(false);
      }
      List<String> items = new ArrayList<>();
      items.add(item);
      /*
       * Progress 10.2B compares items for the RADIO-SET by label (case sensitive) and
       * if one tries to delete an item with label which doesn't exist in a case-sensitive sense 
       * but does exist in a case-insensitive one 
       * it crashes with "SYSTEM ERROR: Memory violation. (49)"
       */
      logical rc = delete(items, COMPARE_LABEL, true);
      if (!rc.booleanValue())
      {
         invalidDeleteArg();
      }
      return rc;
   }

   /**
    * Implements the DELETE() widget method, which removes the list item at the
    * given index.
    * 
    * @param idx
    *           The index of the item to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(int64 idx)
   {
      UnimplementedFeature.unsupported("DELETE() method is undefined");
      return new logical();
   }

   /**
    * Implements the DELETE() widget method, which removes the list item at the
    * given index.
    * 
    * @param idx
    *           The index of the item to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(long idx)
   {
      UnimplementedFeature.unsupported("DELETE() method is undefined");
      return new logical();
   }

   /**
    * Implements the ADD-LAST() widget method, which adds a single label
    * and value pair to the end of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(character label, BaseDataType value)
   {
      if (!canAccess("ADD-LAST"))
      {
         return new logical(false);
      }
     ControlSetItem[] items = config.items;

      if (items == null)
         items = new ControlSetItem[0];
      int currentHeight = items.length;
      int currentWidth = width(items);

      ControlSetItem[] nitems = Arrays.copyOf(items, items.length + 1);
      ControlSetItem item = controlSetItem(label, value);
      nitems[items.length] = item;
      item.setEnabled(!config.horizontal || resetSize(nitems));

      super.setItems(nitems, true);

      return new logical(true);
   }
   

   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   Always reports warning 4052 and returns "unknown".
    */
   @Override
   public logical addFirst(character label, BaseDataType value)
   {
      ErrorManager.recordOrShowError(4052, String.format(
            "ADD-FIRST is not a queryable attribute for %s %s", type(),
               widgetName()), false);
      return new logical();
   }

   /**
    * Store array of items into config.
    * 
    * @param    items
    *           Array of new items.
    */
   @Override
   public void setItems(ControlSetItem[] items, boolean pairs)
   {
      super.setItems(items, true);
      resetSize(items);
   }


   /**
    * Enable RADIO-BUTTON in a RADIO-SET widget. 
    * 
    * @param   label
    *          New label of the button to be enabled.
    *  
    * @return
    *          <code>true</code> if successful
    */
   @Override
   public logical enable(String label)
   {
      return setEnabled(new character(label), true);
   }

   /**
    * Enable RADIO-BUTTON in a RADIO-SET widget. 
    * 
    * @param   label
    *          New label of the button to be enabled.
    *  
    * @return
    *          <code>true</code> if successful
    */
   @Override
   public logical enable(character label)
   {
      return setEnabled(label, true);
   }

   /**
    * Disable RADIO-BUTTON in a RADIO-SET widget. 
    * 
    * @param   label
    *          New label of the button to be disabled.
    *  
    * @return
    *          <code>true</code> if successful
    */
   @Override
   public logical disable(String label)
   {
      return setEnabled(new character(label), false);
   }

   /**
    * Disable RADIO-BUTTON in a RADIO-SET widget. 
    * 
    * @param   label
    *          New label of the button to be disabled.
    *  
    * @return
    *          <code>true</code> if successful
    */
   @Override
   public logical disable(character label)
   {
      return setEnabled(label, false);
   }

   /**
    * Get the value of the NUM-BUTTONS attribute which is the number of buttions in the
    * RADIO-SET.
    *
    * @return   The number of buttons in the widget.
    */
   @Override
   public integer getNumButtons()
   {
      return config.items == null ? new integer(0) : new integer(config.items.length);
   }


   /**
    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character from, character to, integer flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
.    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String from, character to, integer flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
.    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character from, character to, int flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
.    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String from, character to, int flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character from, String to, int flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }


   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String from, String to, integer flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }
   
   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String from, String to, int flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE(REPLACE (new-item-list, list-item )) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character newItemList, character listItem)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
     return missingArgs();
   }

   /**
    * REPLACE(REPLACE (new-item-list, list-item )) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character newItemList, String listItem)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
     return missingArgs();
   }

   /**
    * REPLACE(REPLACE (new-item-list, list-item )) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String newItemList, character listItem)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE(REPLACE (new-item-list, list-item )) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String newItemList, String listItem)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE (new-item-list, list-index ) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character newItemList, integer listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE (new-item-list, list-index ) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character newItemList, int listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE (new-item-list, list-index ) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String newItemList, integer listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE (new-item-list, list-index ) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String newItemList, int listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE (new-label, new-value, list-item) method for COMBO-BOX/SELECTION-LIST,
    * REPLACE (new-label, new-value, old-label) method for RADIO-SET
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param oldItem
    *          oldItem (old-label for the READIO-SET, and list-item for the COMBO-BOX/SELECTION-LIST)
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character newLabel, BaseDataType newValue, character oldItem)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      int idx = getItemIndex(oldItem) - 1;
      if (idx < 0)
      {
         invalidReplaceArg();
         return new logical(false);
      }
      ControlSetItem item = config.items[idx];
      
      character empty = new character("");
      if (empty.equals(newLabel))
      {
         newLabel = item.getLabel();
      }
      
      if (empty.equals(newValue))
      {
         newValue = item.getValue();
      }

      ControlSetItem nitem = controlSetItem(newLabel, newValue);
      config.items[idx] = nitem;
      resetSize(config.items);
      item.setEnabled(item.isEnabled());
      super.setItems(config.items, true);
      
      return new logical(true);
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character newLabel, BaseDataType  newValue, integer listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(String newLabel, BaseDataType newValue, integer listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * 
    * @param arg0
    *          new item's label for COMBO-BOX/SELECTION-LIST,
    *          the source text for EDITOR
    * @param arg1         
    *          new item's value for COMBO-BOX/SELECTION-LIST,
    *          replacement test for editor
    * @param arg2
    *          index of the item to be replaced for COMBO-BOX/SELECTION-LIST,
    *          the search mode for EDITOR
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character arg0, String arg1, integer arg2)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character newLabel, BaseDataType newValue, int listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(String newLabel, BaseDataType  newValue, int listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * Enable/Disable RADIO-BUTTON in a RADIO-SET widget. 
    * 
    * @param   label
    *          New label of the button to be enabled.
    * @param   enabled 
    *          new 'enabled' status
    *          
    * @return
    *          <code>true</code> if successful
    */
   protected logical setEnabled(character label, boolean enabled)
   {
      if (label != null && !label.isUnknown() && config.items != null)
      {
         String l = label.toStringMessage();
         for (int i = 0; i < config.items.length; i++)
         {
            ControlSetItem item = config.items[i];
            if (COMPARE_LABEL_IGNORE_CASE.test(l, item))
            {
               item.setEnabled(enabled);
               pushScreenDefinition();
               return new logical(true);
            }
         }
      }
      ErrorManager.recordOrShowError(4095, String.format(
                  "The radio button in the ENABLE attribute is invalid for %s %s", 
                  type(), widgetName()
               ), 
               false);
      return new logical(false);
   }
 
   /**
    * Calculate widget width for items' list
    * @param items
    *        list of items 
    * @return
    *        the total width of widget associated with the items 
    */
   protected int width(ControlSetItem[] items)
   {
      if (items == null)
      {
         return 0;
      } 
      int maxButtonWidth = 0;
      int totalWidth = 0;
      for (ControlSetItem item: items)
      {
         character label = item.getLabel();
         int lw = label== null ? 0 : label.toStringMessage().length();
         maxButtonWidth = Math.max(maxButtonWidth, lw);
         totalWidth += lw + 3 + 1;
      }
      if (config.expand)
      {
         totalWidth = items.length * (maxButtonWidth + 3 + 1);
      } 
      return config.horizontal ? Math.max(0,totalWidth - 1) : maxButtonWidth;
   }

   /**
    * Replace array of items in the config.
    * 
    * 
    * @param    items
    *           Array of new items.
    * @return
    *           true if size changed OK (or not changed(
    */
   protected boolean resetSize(ControlSetItem[] items)
   {
      if (!config.autoResize)
      {
         return true;
      }

      int height = items == null ? 0 : items.length;
      int width = width(items);
      boolean overflow = false;
      if (dynamic().booleanValue() ? frame != null : config.realized)
      {
         FrameConfig frameConfig = frame.getFrameWidget().config();
         int maxSize = (int)(config.horizontal ? frameConfig.widthChars 
                                               :frameConfig.heightChars);

         if (frameConfig.box) 
         {
            maxSize -= 2;
         }
         
         overflow = maxSize < (config.horizontal ? width : height); 
      }

      if (overflow)
      {
         ErrorManager.recordOrShowError(config.horizontal ? 4475 : 4474, 
               String.format("New %s for %s %s is too large to fit in frame",
                           config.horizontal ? "width" : "height",
                           type(), widgetName()
               ), 
               false);
      }
      else 
      {
         config.listHeightChars = Coordinate.scale(height);
         config.listWidthChars = Coordinate.scale(width);
      }
      return !overflow;
   }

   /**
    * Report invalid argument for delete().
    * 
    */
   private void invalidDeleteArg()
   {
      ErrorManager.recordOrShowError(4068, 
            String.format(
               "The radio button in the DELETE attribute is invalid for %s %s", 
               type(), widgetName()), 
            false);
   }

   /**
    * Report invalid argument for replace().
    * 
    */
   private void invalidReplaceArg()
   {
      ErrorManager.recordOrShowError(4095, 
            String.format(
               "The radio button in the REPLACE attribute is invalid for %s %s", 
               type(), widgetName()), 
            false);
   }

   /**
    * Get item index
    *  
    * @param itemId
    *       item's label or value
    *       
    * @return
    *       item's index or -1 of not found
    */
   @Override
   protected int getItemIndex(character itemId)
   {
      if (itemId == null || itemId.isUnknown())
      {
         return -1;
      }
      return itemIndex(itemId.toStringMessage(), COMPARE_LABEL) + 1;
   }

   /**
    * Report missing argument for replace().
    * 
    */
   private logical missingArgs()
   {
         ErrorManager.recordOrShowError(4064, 
               String.format(
                  "Missing argument when setting attribute REPLACE for %s %s", 
                  type(), widgetName()), 
               false);
         ErrorManager.recordOrShowError(4065, 
               String.format(
                  "The REPLACE attribute on the %s %s has invalid arguments", 
                  type(), widgetName()), 
               false);
         return new logical(false);
   }
      
   /**
    * Report invalid third argument for replace().
    * 
    */
   private logical invalid3RdArg()
   {
      ErrorManager.recordOrShowError(4068, 
            String.format(
               "The third argument for REPLACE on the %s %s should not be of type %s", 
               type(), widgetName(), "INTEGER"), 
            false);
      return new logical(false);
   }
}
