/*
** Module   : SelectionListConfig.java
** Abstract : stores selection list specific state
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050714   @21692 Creation based on general concept. Necessary
**                           modification to get compiled.
** 002 EVL 20050723   @21804 Adding two more attributes: innerChars and
**                           innerLines.
** 003 SIY 20050831   @22514 Refactoring (see @@22394-22409). 
** 004 SIY 20050909   @22670 Fixed initial placement issue.
** 005 NVS 20051116   @23345 Class access changed to public. 
** 006 SIY 20051129   @23564 Moved item list member to parent class.
** 007 SIY 20060209   @24441 Major refactoring.
** 008 GES 20060803   @28416 Made externalizable for better performance.
**                           Many code standard cleanups.
** 009 SIY 20080414   @37972 Removed setColumn() method in order to make 
**                           SelectionList behave like other widgets from
**                           the point of view of ZeroColumnLayout.
** 010 SIY 20100619          Switching to own implementation of character UI.
** 011 SIY 20100804          Changes in imports, some refactorings.
** 012 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 013 SIY 20111202          Updated imports.
** 014 SIY 20111121          Cleaned up imports. 
** 015 HC  20140911          GUI coordinate support - data types refactoring.
** 016 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 017 HC  20141021          Implemented window move/position support.
** 018 GES 20141124          Javadoc fix.
** 019 CA  20150105          Changes for SIDE-LABEL-HANDLE runtime support.
** 020 HC  20150323          Removed business logic from config classes.
** 021 IAS 20150604          Changed default value for the innerLines field.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * Stores selection-list specific state.
 */
public class SelectionListConfig
extends ControlSetConfig
{
   /** The width in chars of the item list. */
   public int innerChars = 10;
   
   /** The number of text lines in the widget. */
   public int innerLines = -1;
   
   /** The operation mode, single or multiple. */
   public boolean multiple = false;
   
   /** Must the list have the horizontal scroll-bar. */
   public boolean scrollbarHorizontal = false;
   
   /** Must the list have the vertical scroll-bar. */
   public boolean scrollbarVertical = false;
   

   /**
    * Default constructor (only used in deserialization).
    */
   public SelectionListConfig()
   {
      possibleSideLabel = true;
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected SelectionListConfig(int id)
   {
      super(id);

      possibleSideLabel = true;
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected SelectionListConfig(WidgetId id)
   {
      super(id);

      possibleSideLabel = true;
   }
   
   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      SelectionListConfig cfg = (SelectionListConfig) config;
      
      innerChars = cfg.innerChars;
      innerLines = cfg.innerLines;
      multiple = cfg.multiple;
      scrollbarHorizontal = cfg.scrollbarHorizontal;
      scrollbarVertical = cfg.scrollbarVertical;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      innerChars          = in.readInt();
      innerLines          = in.readInt();
      multiple            = in.readBoolean();
      scrollbarHorizontal = in.readBoolean();
      scrollbarVertical   = in.readBoolean();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeInt(innerChars);
      out.writeInt(innerLines);
      out.writeBoolean(multiple);
      out.writeBoolean(scrollbarHorizontal);
      out.writeBoolean(scrollbarVertical);
   }
}
